"""
~ Yass ~
"""

import os
import sys
import re
import shutil
import jinja2
import yaml
import json
import logging
import frontmatter
from distutils.dir_util import copy_tree
import markdown
import pkg_resources
import arrow
from jinja_macro_tags import configure_environment
from webassets.ext.jinja2 import AssetsExtension
from webassets.script import CommandLineEnvironment as WACommandLineEnvironment
from webassets import Environment as WAEnv
import webassets.loaders
import pyjade
from slugify import slugify


# ------------------------------------------------------------------------------

NAME = "Yass"

PAGE_FORMAT = (".html", ".md", ".jade")

DEFAULT_LAYOUT = "layouts/default.html"


# Markdown
mkd = markdown.Markdown(extensions=[
    'markdown.extensions.nl2br',
    'markdown.extensions.sane_lists',
    'markdown.extensions.toc',
    'markdown.extensions.tables'
])


def md_to_html(text):
    """
    Convert MD text to HTML
    :param text:
    :return:
    """
    html = mkd.convert(text)
    mkd.reset()
    return html


def md_get_toc(text):
    """
    Extract Table of Content of MD
    :param text:
    :return:
    """
    mkd.convert(text)
    toc = mkd.toc
    mkd.reset()
    return toc


class dictdot(dict):
    """
    A dict extension that allows dot notation to access the data.
    ie: dict.get('key.key2.0.keyx'). Still can use dict[key1][k2]
    To create: dictdot(my)
    """
    def get(self, key, default=None):
        """ access data via dot notation """
        try:
            val = self
            if "." not in key:
                return self[key]
            for k in key.split('.'):
                if k.isdigit():
                    k = int(k)
                val = val[k]
            return val
        except (TypeError, KeyError, IndexError) as e:
            return default


def load_conf(yml_file):
    with open(yml_file) as f:
        return dictdot(yaml.load(f))


def format_datetime(dt, format):
    return "" if not dt else arrow.get(dt).format(format)


def jade_to_html(text):
    return pyjade.simple_convert(text)


# ==============================================================================
# -------------------------------- YASS ----------------------------------------
# ==============================================================================

class Yass(object):

    RE_BLOCK_TITLE = re.compile(r'{%\s*block\s+title\s*%}')
    RE_BLOCK_TITLE_PARSED = re.compile(r'{%\s*block\s+title\s*%}(.*?){%\s*endblock\s*%}')
    RE_BLOCK_BODY = re.compile(r'{%\s*block\s+body\s*%}')
    RE_BLOCK_BODY_PARSED = re.compile(r'{%\s*block\s+body\s*%}(.*?){%\s*endblock\s*%}')
    RE_EXTENDS = re.compile(r'{%\s*extends\s+(.*?)\s*%}')

    # Global context
    context = {
        "site": {},     # context set in the yass.yml
        "page": {},     # context set at the page level
        "data": {}      # context set in the data directory
    }

    default_page_meta = {
        "title": "",            # The title of the page
        "markup": None,         # The markup to use. ie: md | jade | html (default)
        "slug": None,           # The pretty url new name of the file. A file with the same name will be created
        "url": None,            # This will be added when processed
        "description": "",      # Page description
        "pretty_url": True,     # By default, all url will be pretty (search engine friendly) Set to False to keep the .html
        "meta": {}
    }

    def __init__(self, root_dir):
        self.root_dir = root_dir
        self.build_dir = os.path.join(self.root_dir, "build")
        self.static_dir = os.path.join(self.root_dir, "static")
        self.content_dir = os.path.join(self.root_dir, "content")
        self.pages_dir = os.path.join(self.root_dir, "pages")
        self.templates_dir = os.path.join(self.root_dir, "templates")
        self.data_dir = os.path.join(self.root_dir, "data")
        self.build_static_dir = os.path.join(self.build_dir, "static")

        config_file = os.path.join(self.root_dir, "yass.yml")
        self.config = load_conf(config_file)
        self.default_layout = self.config.get("default_layout", DEFAULT_LAYOUT)

        # Site context
        self.context["site"] = self.config.get("site", {})
        if "base_url" not in self.context["site"] or not self.context["site"]["base_url"]:
            self.context["site"]["base_url"] = "/"
        self.base_url = self.context["site"]["base_url"]

        self.sitename = self.config.get("sitename")
        if self.sitename:
            self.sitename = self.sitename.lstrip("http://")\
                .lstrip("https://")\
                .lstrip("www.")



    def setup(self):
        if not os.path.isdir(self.build_dir):
            os.makedirs(self.build_dir)
        self.setup_jinja()
        self.setup_webassets()

    def setup_jinja(self):
        # Jinja

        loader = jinja2.ChoiceLoader([
            # global macros
            jinja2.DictLoader({
                "yass.macros": pkg_resources.resource_string(__name__,
                                                             "macros.html"),
            }),
            jinja2.FileSystemLoader(self.templates_dir)
        ])

        self.tpl_env = jinja2.Environment(loader=loader,
                                          extensions=[
                                              'yass.htmlcompress.HTMLCompress',
                                              'pyjade.ext.jinja.PyJadeExtension',
                                               AssetsExtension])

        configure_environment(self.tpl_env)
        self.tpl_env.filters.update({
            "format_datetime": format_datetime,
            "markdown_to_html": md_to_html,
            "jade_to_html": jade_to_html
        })
        self.tpl_env.macros.register_from_template('yass.macros')

    def setup_webassets(self):

        assets_env = WAEnv(directory="./static",
                           url=self.config.get("static_url", "/static"))
        bundles = self.config.get("assets_bundles", {})
        assets_env.register(bundles)
        self.tpl_env.assets_environment = assets_env

        self.webassets_cmd = None
        if bundles:
            handler = logging.StreamHandler if self.config.get("debug", False) \
                else logging.NullHandler
            log = logging.getLogger('webassets')
            log.addHandler(handler())
            log.setLevel(logging.DEBUG)
            self.webassets_cmd = WACommandLineEnvironment(assets_env, log)

    def clean_build_dir(self):
        if os.path.isdir(self.build_dir):
            shutil.rmtree(self.build_dir)
        os.makedirs(self.build_dir)

    def load_data(self):
        # Data context
        self.context["data"] = {}
        for root, _, files in os.walk(self.data_dir):
            for fname in files:
                if fname.endswith((".yml", ".json")):
                    name = fname.replace(".yml", "").replace(".json", "")
                    fname = os.path.join(root, fname)
                    if os.path.isfile(fname):
                        with open(fname) as f:
                            _ = {}
                            if fname.endswith(".yml"):
                                _ = yaml.load(f)
                            elif fname.endswith(".json"):
                                _ = json.load(f)
                            self.context["data"][name] = _

    def set_context(self, key, val):
        """
        To programmatically add new context
        :param key:
        :param val:
        :return:
        """
        self.context[key] = val

    def build_static(self):
        """
        Build static assets on demand
        :return:
        """
        if not os.path.isdir(self.build_static_dir):
            os.makedirs(self.build_static_dir)

        copy_tree(self.static_dir, self.build_static_dir)

        if self.webassets_cmd:
            self.webassets_cmd.build()

    def build_pages(self):
        """
        Build all the pages
        :return:
        """
        for root, _, files in os.walk(self.pages_dir):
            base_dir = root.replace(self.pages_dir, "").lstrip("/")
            dest_dir = (self.build_dir + "/" + base_dir).rstrip("/")
            if not base_dir.startswith("_"):
                for f in files:
                    src_file = root + "/" + f
                    self._build_page(dest_dir=dest_dir, src_file=src_file)

    def _build_page(self, src_file, dest_dir):

        filename = src_file.split("/")[-1]
        # If filename starts with _ (underscore) do not build
        if not filename.startswith("_") and (filename.endswith(PAGE_FORMAT)):
            dest_dir = dest_dir.rstrip("/")
            metadata = self.default_page_meta.copy()
            metadata["meta"].update(self.config.get("site.meta", {}))
            with open(src_file) as f:
                _meta, content = frontmatter.parse(f.read())
                metadata.update(_meta)

            fname = filename\
                .replace(".html", "")\
                .replace(".md", "")\
                .replace(".jade", "")

            pretty_url = metadata.get("pretty_url", True)
            slug = metadata.get("slug")
            if slug:
                fname = slugify(slug)

            if not pretty_url:
                dest_file = os.path.join(dest_dir, "%s.html" % fname)

            else:
                if filename not in ["index.html", "index.md", "index.jade"]:
                    dest_dir = os.path.join(dest_dir, fname)
                dest_file = os.path.join(dest_dir, "index.html")

            if not os.path.isdir(dest_dir):
                os.makedirs(dest_dir)

            # The final url
            metadata["url"] = dest_file.replace(self.build_dir, "")\
                .replace("index.html", "")

            # page context
            context = self.context.copy()
            context["page"] = metadata

            if src_file.endswith(".md") or metadata["markup"] == "md":
                context["page"]["__toc__"] = md_get_toc(content)
                content = md_to_html(content)
            elif src_file.endswith(".jade") or metadata["markup"] == "jade":
                content = jade_to_html(content)

            # Page must be extended by a layout and have a block 'body'
            # These tags will be included if they are missing
            # .jade and .md files don't require these tags

            if re.search(self.RE_EXTENDS, content) is None:
                layout = metadata["layout"] if "layout" in metadata else self.default_layout
                content = ("\n{% extends '{}' %} \n\n"
                           .replace("{}", layout)) + content

            if re.search(self.RE_BLOCK_BODY, content) is None:
                _layout_block = re.search(self.RE_EXTENDS, content).group(0)
                content = content.replace(_layout_block, "")
                content = "\n" + _layout_block + "\n" + \
                          "{% block body %} \n" + content.strip() + "\n{% endblock %}"

            tpl = self.tpl_env.from_string(content)
            with open(dest_file, "w") as fw:
                fw.write(tpl.render(context))

    def build(self):
        self.clean_build_dir()
        self.setup()
        self.load_data()
        self.build_static()
        self.build_pages()






import os
import pkg_resources
import click
from livereload import Server, shell
from . import Yass, publisher

CWD = os.getcwd()


def copy_resource(src, dest):
    """
    To copy package data to destination
    """
    package_name = "yass"
    dest = (dest + "/" + os.path.basename(src)).rstrip("/")
    if pkg_resources.resource_isdir(package_name, src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        for res in pkg_resources.resource_listdir(__name__, src):
            copy_resource(src + "/" + res, dest)
    else:
        if not os.path.isfile(dest) \
                and os.path.splitext(src)[1] not in [".pyc"]:
            with open(dest, "wb") as f:
                f.write(pkg_resources.resource_string(__name__, src))
        else:
            print("File exists: %s " % dest)

def footer():
    print("-" * 80)

@click.group()
def cli():
    """
    Yass: Static site generator for us!
    """
    pass


@cli.command("build")
def build():
    """Build everything"""
    print("Building pages...")
    Yass(CWD).build()
    print("Done!")

    footer()


@cli.command("publish")
@click.argument("endpoint", default="s3")
def publish(endpoint):
    """Publish the site"""
    print("Publishing site to %s ..." % endpoint.upper())

    yass = Yass(CWD)
    target = endpoint.lower()

    sitename = yass.sitename
    if not sitename:
        raise ValueError("Missing site name")

    endpoint = yass.config.get("publish_endpoints.%s" % target)
    if not endpoint:
        raise ValueError("%s endpoint is missing in the config" % target.upper())

    if target == "s3":
        p = publisher.S3Website(sitename=sitename,
                                aws_access_key_id=endpoint.get("aws_access_key_id"),
                                aws_secret_access_key=endpoint.get("aws_secret_access_key"),
                                region=endpoint.get("aws_region"))

        if not p.website_exists:
            print("Setting S3 site...")
            if p.create_website() is True:
                p.create_www_website()
                print("New bucket created: %s" % p.sitename)

        print("Uploading pages and static files...")
        p.upload(yass.build_dir)

        print("")
        print("Yass! Your site has been successfully published to: ")
        print(p.website_endpoint_url)

    footer()


@cli.command("setup-domain")
@click.argument("endpoint", default="s3")
def setup_domain(endpoint):
    """Setup site domain to route to static site"""
    print("Setting up domain...")

    yass = Yass(CWD)
    target = endpoint.lower()

    sitename = yass.sitename
    if not sitename:
        raise ValueError("Missing site name")

    endpoint = yass.config.get("publish_endpoints.%s" % target)
    if not endpoint:
        raise ValueError(
            "%s endpoint is missing in the config" % target.upper())

    if target == "s3":
        p = publisher.S3Website(sitename=sitename,
                                aws_access_key_id=endpoint.get("aws_access_key_id"),
                                aws_secret_access_key=endpoint.get("aws_secret_access_key"),
                                region=endpoint.get("aws_region"))

        print("Setting AWS Route53 for: %s ..." % p.sitename)
        p.setup_domain()
        print("")
        print("Yass! Route53 setup successfully!")
        print("You can now visit the site at :")
        print(p.sitename_endpoint)
    footer()


@cli.command("create")
@click.argument("sitename")
def create(sitename):
    """Create a new site."""
    sitepath = os.path.join(CWD, sitename)
    if os.path.isdir(sitepath):
        print("Site directory '%s' exists already!" % sitename)
    else:
        print("Creating site: %s..." % sitename)
        os.makedirs(sitepath)
        copy_resource("skel/", sitepath)
        print("Site created successfully!")
        print("CD into '%s' and run 'yass serve' to view the site" % sitename)

    footer()




@cli.command()
@click.option("-p", "--port", default=None)
@click.option("--no-livereload", default=None)
@click.option("--open-url", default=None)
def serve(port, no_livereload, open_url):
    """Serve the site """

    engine = Yass(CWD)
    if not port:
        port = engine.config.get("local_server.port", 8000)
    if no_livereload is None:
        no_livereload = True if engine.config.get("local_server.livereload") is False else False
    if open_url is None:
        open_url = False if engine.config.get("local_server.open_url") is False else True

    print("Serving at %s" % port)
    print("Livereload is %s" % ("OFF" if no_livereload else "ON"))

    def build_static():
        engine.build_static()

    def build_pages():
        engine.build_pages()

    engine.build()

    server = Server()
    if no_livereload is False:
        server.watch(engine.static_dir + "/", build_static)
        server.watch(engine.pages_dir + "/", build_pages)
        server.watch(engine.templates_dir + "/", build_pages)
        server.watch(engine.data_dir + "/", build_pages)

    server.serve(open_url_delay=open_url, port=port, root=engine.build_dir)


@cli.command("clean")
def clean():
    """Clean the build dir """
    print("Cleaning build dir...")
    Yass(CWD).clean_build_dir()
    print("Done!")
    footer()


def cmd():
    try:
        print("=" * 80)
        print("Yass!")
        print("-" * 80)
        cli()
    except Exception as e:
        print("Ohhh noooooo! Something bad happens")
        print(">> %s " % e)
        raise e





import botocore
import boto3
import json
import os
import mimetypes
import threading
import uuid

MIMETYPE_MAP = {
    '.js':   'application/javascript',
    '.mov':  'video/quicktime',
    '.mp4':  'video/mp4',
    '.m4v':  'video/x-m4v',
    '.3gp':  'video/3gpp',
    '.woff': 'application/font-woff',
    '.woff2': 'font/woff2',
    '.eot':  'application/vnd.ms-fontobject',
    '.ttf':  'application/x-font-truetype',
    '.otf':  'application/x-font-opentype',
    '.svg':  'image/svg+xml',
}
MIMETYPE_DEFAULT = 'application/octet-stream'


def get_mimetype(filename):
    mimetype, _ = mimetypes.guess_type(filename)
    if mimetype:
        return mimetype

    base, ext = os.path.splitext(filename)
    ext = ext.lower()
    if ext in MIMETYPE_MAP:
        return MIMETYPE_MAP[ext]
    return MIMETYPE_DEFAULT


class S3Website(object):
    """
    To manage S3 website and domain on Route53
    """
    S3_HOSTED_ZONE_IDS = {
        'us-east-1': 'Z3AQBSTGFYJSTF',
        'us-west-1': 'Z2F56UZL2M1ACD',
        'us-west-2': 'Z3BJ6K6RIION7M',
        'ap-south-1': 'Z11RGJOFQNVJUP',
        'ap-northeast-1': 'Z2M4EHUR26P7ZW',
        'ap-northeast-2': 'Z3W03O7B5YMIYP',
        'ap-southeast-1': 'Z3O0J2DXBE1FTB',
        'ap-southeast-2': 'Z1WCIGYICN2BYD',
        'eu-central-1': 'Z21DNDUVLTQW6Q',
        'eu-west-1': 'Z1BKCTXD74EZPE',
        'sa-east-1': 'Z7KQH4QJS55SO',
        'us-gov-west-1': 'Z31GFT0UA1I2HV',
    }

    def __init__(self,
                 sitename,
                 region="us-east-1",
                 aws_access_key_id=None,
                 aws_secret_access_key=None):
        """

        :param sitename: the website name to create, without WWW.
        :param region: the region of the site
        :param access_key_id: AWS
        :param secret_access_key: AWS
        :param setup_dns: bool - If True it will create route53
        :param allow_www: Bool - If true, it will create a second bucket with www.
        """

        # This will be used to pass to concurrent upload
        self.aws_params = {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "region_name": region
        }
        self.region = region
        self.s3 = boto3.client('s3', **self.aws_params)
        self.sitename = sitename.lstrip("www.")
        self.www_sitename = "www." + self.sitename
        self.website_endpoint = "%s.s3-website-%s.amazonaws.com" % (self.sitename, region)
        self.website_endpoint_url = "http://" + self.website_endpoint
        self.sitename_endpoint = "http://" + self.sitename
        exists, error_code, error_message = self.head_bucket(self.sitename)
        self.website_exists = exists

    def setup_domain(self):
        route53 = boto3.client('route53', **self.aws_params)

        hosted_zone_id = self._get_route53_hosted_zone_by_domain(self.sitename)
        if not hosted_zone_id:
            caller_reference_uuid = "%s" % (uuid.uuid4())
            response = route53.create_hosted_zone(
                Name=self.sitename,
                CallerReference=caller_reference_uuid,
                HostedZoneConfig={'Comment': "HostedZone created by YASS!", 'PrivateZone': False})
            hosted_zone_id = response['HostedZone']['Id']

        website_dns_name = "s3-website-%s.amazonaws.com" % self.region
        redirect_dns_name = "s3-website-%s.amazonaws.com" % self.region

        change_batch_payload = {
            'Changes': [
                {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'Name': self.sitename,
                        'Type': 'A',
                        'AliasTarget': {
                            'HostedZoneId': self.S3_HOSTED_ZONE_IDS[self.region],
                            'DNSName': website_dns_name,
                            'EvaluateTargetHealth': False
                        }
                    }
                },
                {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'Name': self.www_sitename,
                        'Type': 'A',
                        'AliasTarget': {
                            'HostedZoneId': self.S3_HOSTED_ZONE_IDS[self.region],
                            'DNSName': redirect_dns_name,
                            'EvaluateTargetHealth': False
                        }
                    }
                }
            ]
        }

        response = route53.change_resource_record_sets(
            HostedZoneId=hosted_zone_id,
            ChangeBatch=change_batch_payload)
        return True if response and "ChangeInfo" in response else False

    def _get_route53_hosted_zone_by_domain(self, domain):
        route53 = boto3.client('route53', **self.aws_params)
        hosted_zone = route53.list_hosted_zones()

        if hosted_zone or "HostedZones" in hosted_zone:
            for hz in hosted_zone["HostedZones"]:
                if hz["Name"].rstrip(".") == domain:
                    return hz["Id"]
        return None



    def head_bucket(self, name):
        """
        Check if a bucket exists
        :param name:
        :return:
        """
        try:
            self.s3.head_bucket(Bucket=name)
            info = self.s3.get_bucket_website(Bucket=self.sitename)
            if not info:
                return False, 404, "Configure improrperly"
            return True, None, None
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] in ["403", "404"]:
                return False, e.response["Error"]["Code"], e.response["Error"]["Message"]
            else:
                raise e

    def create_website(self):
        exists, error_code, error_message = self.head_bucket(self.sitename)
        if not exists:
            if error_code == "404":
                # Allow read access
                policy_payload = {
                    "Version": "2012-10-17",
                    "Statement": [{
                        "Sid": "Allow Public Access to All Objects",
                        "Effect": "Allow",
                        "Principal": "*",
                        "Action": "s3:GetObject",
                        "Resource": "arn:aws:s3:::%s/*" % (self.sitename)
                    }
                    ]
                }
                # Make bucket website and add index.html and error.html
                website_payload = {
                    'ErrorDocument': {
                        'Key': 'error.html'
                    },
                    'IndexDocument': {
                        'Suffix': 'index.html'
                    }
                }
                self.s3.create_bucket(Bucket=self.sitename)
                self.s3.put_bucket_policy(Bucket=self.sitename,
                                          Policy=json.dumps(policy_payload))
                self.s3.put_bucket_website(Bucket=self.sitename,
                                           WebsiteConfiguration=website_payload)
                return True
            else:
                raise Exception("Can't create website's bucket '%s' on AWS S3. "
                                "Error: %s" % (self.sitename, error_message))
        return False

    def create_www_website(self):
        exists, error_code, error_message = self.head_bucket(self.sitename)
        if not exists:
            if error_code == "404":
                self.s3.create_bucket(Bucket=self.www_sitename)
                redirect_payload = {
                    'RedirectAllRequestsTo': {
                        'HostName': self.sitename,
                        'Protocol': 'http'
                    }
                }
                bucket_website_redirect = self.s3.BucketWebsite(self.www_sitename)
                bucket_website_redirect.put(WebsiteConfiguration=redirect_payload)
                return True
            else:
                raise Exception("Can't create website's bucket '%s' on AWS S3. "
                                "Error: %s" % (self.www_sitename, error_message))
        return False

    def upload(self, build_dir):

        for root, dirs, files in os.walk(build_dir):
            for filename in files:
                local_path = os.path.join(root, filename)
                s3_path = os.path.relpath(local_path, build_dir)
                mimetype = get_mimetype(local_path)

                kwargs = dict(aws_params=self.aws_params,
                              bucket_name=self.sitename,
                              local_path=local_path,
                              s3_path=s3_path,
                              mimetype=mimetype)

                threading.Thread(target=self._upload_file, kwargs=kwargs)\
                    .start()

    @staticmethod
    def _upload_file(aws_params, bucket_name, local_path, s3_path, mimetype):
        s3 = boto3.client("s3", **aws_params)
        s3.upload_file(local_path,
                       Bucket=bucket_name,
                       Key=s3_path,
                       ExtraArgs={"ContentType": mimetype})




"""
Yass!

Yet Another Static Site!

"""

import os
from setuptools import setup, find_packages

__version__ = "0.0.5"

install_requires = [
    "jinja2==2.8",
    "webassets==0.12.0",
    "click==6.2",
    "pyyaml==3.11",
    "markdown==2.6.2",
    "pyjade==4.0.0",
    "python-frontmatter==0.3.1",
    "jinja-macro-tags==0.1",
    "livereload==2.5.0",
    "arrow==0.8.0",
    "python-slugify==1.2.1",
    "boto3==1.4.3",
]

setup(
    name="Yass",
    version=__version__,
    license="MIT",
    author="Mardix",
    author_email="mardix@pylot.io",
    description="Yet Another Static Site generator for everybody",
    url="https://github.com/mardix/yass",
    long_description="Static site generator",
    py_modules=['yass'],
    entry_points=dict(console_scripts=[
        'yass=yass.cli:cmd'
    ]),
    include_package_data=True,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=install_requires,
    keywords=['static site generator'],
    platforms='any',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe=False
)